#include <asm/cpuid.h>
#include <asm/msr.h>
#include <asm/page.h>
#include <asm/page_types.h>
#include <asm/paravirt.h>
#include <linux/errno.h>
#include <linux/gfp_types.h>
#include <linux/init.h> /* Needed for the macros */
#include <linux/kern_levels.h>
#include <linux/module.h> /* Needed by all modules */
#include <linux/printk.h>
#include <linux/slab.h>

/*============== my includes ==============*/
#include "asm/smp.h"
#include "asm/special_insns.h"
#include "asm/tlbflush.h"
#include "debug/debug.h"
#include "linux/threads.h"
#include "msr/msr.h"
#include "region/vxmon.h"
#include "vcpu/vcpu.h"

#define NULL ((void *)0)

/*asm ( assembler template*/
/*    : output operands                   (optional)*/
/*    : input operands                    (optional)*/
/*    : clobbered registers list          (optional)*/
/*    );*/

static struct vcpu_t vcpus[1];

static int cr4_enable_vmx(void) {
    unsigned long cr4;

    /*BUG: Need to somehow switch to a CPU whose VMXE bit hasn't be modified*/
    /*NOTE: seems like this bit is always set on every processor, weird..*/
    /*if ((__read_cr4() >> 13) & 1) {*/
    /*    return -EBUSY;*/
    /*}*/

    __asm__ volatile("mov %%cr4, %0" : "=r"(cr4));
    cr4 |= (1UL << 13);
    __asm__ volatile("mov %0, %%cr4" ::"r"(cr4));

    DEBUG_FMT("CR4[13].VMXE set? %s\n",
              (__read_cr4() >> 13) & 1 ? "true" : "false");

    return 0;
}

static bool vmx_support_cpuid(void) {
    unsigned int eax, ebx, ecx, edx;
    cpuid(0, &eax, &ebx, &ecx, &edx);

    unsigned int ret_str[3] = {ebx, edx, ecx};

    DEBUG_FMT("Vendor ID = %s\n", (char *)ret_str);

    unsigned int ecx_vmx = cpuid_ecx(1);

    /*__asm__ volatile("mov $1, %rax");*/
    /*__asm__ volatile("cpuid");*/
    /*__asm__ volatile("mov %%ecx , %0\n\t" : "=r"(ecx));*/

    unsigned int addr_width = cpuid_eax(0x80000008);

    DEBUG_FMT("Physical address width = %d\n", addr_width & 0xff);

    return (ecx_vmx >> 5) & 1;
}

/*
 * https://elixir.bootlin.com/linux/v6.12.4/source/tools/testing/selftests/kvm/include/x86_64/vmx.h#L297
 */
static unsigned char vmxon(unsigned long long pa) {
    unsigned char ret;

    // setna: "Set byte if not above" (CF=1 or ZF=1)
    __asm__ volatile("vmxon %[pa]; setna %[ret]"
                     : [ret] "=rm"(ret)
                     : [pa] "m"(pa)
                     : "cc", "memory");

    /*asm goto("1: vmxon %[vmxon_pointer]\n\t" _ASM_EXTABLE(1b, % l[fault])*/
    /*         :*/
    /*         : [vmxon_pointer] "m"(pa)*/
    /*         :*/
    /*         : fault);*/

    unsigned long long rflags;

    __asm__ volatile("pushfq; popq %0" : "=r"(rflags));
    DEBUG_FMT("RFLAGS: 0x%llx\n", rflags);

    return ret;

    /*    unsigned long long msr;*/
    /**/
    /*    cr4_set_bits(X86_CR4_VMXE);*/
    /**/
    /*    __asm__ goto("1: vmxon %[vmxon_pointer]\n\t" _ASM_EXTABLE(1b,
     * %l[fault])*/
    /*                 :*/
    /*                 : [vmxon_pointer] "m"(vmxon_pointer)*/
    /*                 :*/
    /*                 : fault);*/
    /*    return 0;*/
    /**/
    /*fault:*/
    /*    WARN_ONCE(1, "VMXON faulted, MSR_IA32_FEAT_CTL (0x3a) = 0x%llx\n",*/
    /*              rdmsrl_safe(MSR_IA32_FEAT_CTL, &msr) ? 0xdeadbeef : msr);*/
    /*    cr4_clear_bits(X86_CR4_VMXE);*/
    /**/
    /*    return -EFAULT;*/
}

/*NOTE: arch/x86/kvm/vmx/vmx.c:2824 v6.12.10*/
static int kvm_cpu_vmxon(unsigned long long vmxon_pointer) {
    unsigned long long msr;

    cr4_set_bits(X86_CR4_VMXE);

    // clang-format off
    asm goto("1: vmxon %[vmxon_pointer]\n\t" _ASM_EXTABLE(1b, %l[fault])
             :
             : [vmxon_pointer] "m"(vmxon_pointer)
             :
             : fault);

    // clang-format on
    return 0;

fault:
    WARN_ONCE(1, "VMXON faulted, MSR_IA32_FEAT_CTL (0x3a) = 0x%llx\n",
              rdmsrl_safe(MSR_IA32_FEAT_CTL, &msr) ? 0xdeadbeef : msr);
    cr4_clear_bits(X86_CR4_VMXE);

    return -EFAULT;
}

/*NOTE: arch/x86/kvm/vmx/vmx.c:742 v6.12.10*/
static int kvm_cpu_vmxoff(void) {
    // clang-format off
    asm goto("1: vmxoff\n\t" _ASM_EXTABLE(1b, %l[fault])::
                 : "cc", "memory"
             : fault);

    // clang-format on
    cr4_clear_bits(X86_CR4_VMXE);
    return 0;

fault:
    cr4_clear_bits(X86_CR4_VMXE);
    return -EIO;
}

static int my_init(void) {
    /*
     * https://www.sciencedirect.com/topics/computer-science/current-privilege-level
     */
    DEBUG_FMT("Querying CS (lowest 2 bits) for CPL\n");
    unsigned long cs, cpl;

    __asm__ volatile("mov %%cs, %0" : "=r"(cs));
    cpl = cs & 0x3;

    DEBUG_FMT("CPL level = %lu\n", cpl);
    if (cpl > 0) {
        return -1;
    }

    pr_info("Checking VMX support using CPUID\n");
    if (!vmx_support_cpuid()) {
        pr_err("VMX isn't supported\n");
        return -EPERM;
    }

    DEBUG_FMT("IA32_VMX_BASIC_MSR = 0x%llx\n", __rdmsr(IA32_VMX_BASIC));

    pr_info("Allocating VMXON region\n");

    struct vmxon_t *vmxon_reg = alloc_vmxon();

    if (!vmxon_reg) {
        pr_err("VMXON region allocation failed\n");
        return -ENOMEM;
    }

    if (init_vcpu(&vcpus[0], vmxon_reg) != 0) {
        pr_err("VCPU initialization failed\n");
        return -EFAULT;
    }

    pr_info("VA of the allocated region = 0x%px\n", vmxon_reg);
    pr_info("PA of the allocated region = 0x%lx\n", __pa(vmxon_reg));

    pr_info("Reading VMXON region for VMCS ID: 0x%x\n",
            vmxon_reg->header.vmcs_rev_id);

    pr_info("Patching CR0 and CR4 depending on the value of their respective "
            "MSRs\n");
    patch_control_registers();

    pr_info("Enabling VMX in CR4\n");
    int ret_cr4_vmx;

    if ((ret_cr4_vmx = cr4_enable_vmx()) != 0) {
        pr_err("CPU not available, VMXE bit in CR4 has already been set\n");
        return ret_cr4_vmx;
    }

    pr_info("Checking the necessary flags of the IA32_FEATURE_CONTROL_MSR\n");
    if (!ia32_feature_control_flags()) {
        pr_err("The flags of the IA32_FEATURE_CONTROL MSR do not permit "
               "virtualization\n");
        return -EPERM;
    }

    pr_info("Executing VMXON with address = 0x%lx as its operand\n",
            __pa(vmxon_reg));

    unsigned char vmxon_ret = 0;

    /*if ((vmxon_ret = vmxon(vmxon_region.pa) != 0)) {*/
    if ((vmxon_ret = kvm_cpu_vmxon(__pa(vmxon_reg)) != 0)) {
        /*unsigned long vm_err = __rdmsr(0x4400);*/
        /*pr_err("VM_ERR val = 0x%lx\n", vm_err);*/
        cr4_clear_bits(13);

        kfree(vmxon_reg);
        /*__asm__ volatile("vmxoff");*/
        pr_err("`vmxon` failed with return code %d\n", vmxon_ret);
        return -1;
    }

    pr_info("`vmxon` was successfully executed!\n");

    return 0;
}

static void my_exit(void) {
    pr_info("Executing VMXOFF\n");
    /*__asm__ volatile("vmxoff");*/
    int vmxoff;

    if ((vmxoff = kvm_cpu_vmxoff()) != 0) {
        pr_err("Failed to execute VMXOFF\n");
        return;
    }

    pr_info("Freeing memory of the VMXON region\n");
    kfree(vcpus[0].vmxon);

    pr_info("vmbr.ko has exited\n");
}

module_init(my_init);
module_exit(my_exit);

MODULE_LICENSE("GPL");
